from flask_debugtoolbar.panels import DebugPanel
from flask import current_app

_ = lambda x: x


class RouteListDebugPanel(DebugPanel):
    """
    Panel that displays the URL routing rules.
    """
    name = 'RouteList'
    has_content = True
    routes = []

    def nav_title(self):
        return _('Route List')

    def title(self):
        return _('Route List')

    def url(self):
        return ''

    def nav_subtitle(self):
        count = len(self.routes)
        return '%s %s' % (count, 'route' if count == 1 else 'routes')

    def process_request(self, request):
        self.routes = [
            rule
            for rule in current_app.url_map.iter_rules()
            if not rule.rule.startswith('/_debug_toolbar')
        ]

    def content(self):
        return self.render('panels/route_list.html', {
            'routes': self.routes,
        })



from flask import session

from flask_debugtoolbar.panels import DebugPanel

_ = lambda x: x


class RequestVarsDebugPanel(DebugPanel):
    """
    A panel to display request variables (POST/GET, session, cookies).
    """
    name = 'RequestVars'
    has_content = True

    def nav_title(self):
        return _('Request Vars')

    def title(self):
        return _('Request Vars')

    def url(self):
        return ''

    def process_request(self, request):
        self.request = request
        self.session = session
        self.view_func = None
        self.view_args = []
        self.view_kwargs = {}

    def process_view(self, request, view_func, view_kwargs):
        self.view_func = view_func
        self.view_kwargs = view_kwargs

    def content(self):
        context = self.context.copy()
        context.update({
            'get': self.request.args.lists(),
            'post': self.request.form.lists(),
            'cookies': self.request.cookies.items(),
            'view_func': ('%s.%s' % (self.view_func.__module__,
                                     self.view_func.__name__)
                          if self.view_func else '[unknown]'),
            'view_args': self.view_args,
            'view_kwargs': self.view_kwargs or {},
            'session': self.session.items(),
        })

        return self.render('panels/request_vars.html', context)



import ntpath
import posixpath

from flask import Markup
import pytest

from flask_debugtoolbar.utils import (_relative_paths, _shortest_relative_path,
                                      format_sql, decode_text, HAVE_PYGMENTS)


@pytest.mark.parametrize('value,paths,expected,path_module', [
    # should yield relative path to the parent directory
    ('/foo/bar', ['/foo'], ['bar'], posixpath),
    ('c:\\foo\\bar', ['c:\\foo'], ['bar'], ntpath),

    # should not yield result if no path is a parent directory
    ('/foo/bar', ['/baz'], [], posixpath),
    ('c:\\foo\\bar', ['c:\\baz'], [], ntpath),

    # should only yield relative paths for parent directories
    ('/foo/bar', ['/foo', '/baz'], ['bar'], posixpath),
    ('c:\\foo\\bar', ['c:\\foo', 'c:\\baz'], ['bar'], ntpath),

    # should yield all results when multiple parents match
    ('/foo/bar/baz', ['/foo', '/foo/bar'], ['bar/baz', 'baz'], posixpath),
    ('c:\\foo\\bar\\baz', ['c:\\foo', 'c:\\foo\\bar'],
        ['bar\\baz', 'baz'], ntpath),

    # should ignore case differences on windows
    ('c:\\Foo\\bar', ['c:\\foo'], ['bar'], ntpath),

    # should preserve original case
    ('/Foo/Bar', ['/Foo'], ['Bar'], posixpath),
    ('c:\\Foo\\Bar', ['c:\\foo'], ['Bar'], ntpath),
])
def test_relative_paths(value, paths, expected, path_module):
    assert list(_relative_paths(value, paths, path_module)) == expected


@pytest.mark.parametrize('value,paths,expected,path_module', [
    # should yield relative path to the parent directory
    ('/foo/bar', ['/foo'], 'bar', posixpath),
    ('c:\\foo\\bar', ['c:\\foo'], 'bar', ntpath),

    # should return the original value if no path is a parent directory
    ('/foo/bar', ['/baz'], '/foo/bar', posixpath),
    ('c:\\foo\\bar', ['c:\\baz'], 'c:\\foo\\bar', ntpath),

    # should yield shortest result when multiple parents match
    ('/foo/bar/baz', ['/foo', '/foo/bar'], 'baz', posixpath),
    ('c:\\foo\\bar\\baz', ['c:\\foo', 'c:\\foo\\bar'], 'baz', ntpath),
])
def test_shortest_relative_path(value, paths, expected, path_module):
    assert _shortest_relative_path(value, paths, path_module) == expected


def test_decode_text_unicode():
    value = u'\uffff'
    decoded = decode_text(value)
    assert decoded == value


def test_decode_text_ascii():
    value = 'abc'
    assert decode_text(value.encode('ascii')) == value


def test_decode_text_non_ascii():
    value = b'abc \xff xyz'
    assert isinstance(value, bytes)

    decoded = decode_text(value)
    assert not isinstance(decoded, bytes)

    assert decoded.startswith('abc')
    assert decoded.endswith('xyz')


@pytest.fixture()
def no_pygments(monkeypatch):
    monkeypatch.setattr('flask_debugtoolbar.utils.HAVE_PYGMENTS', False)


def test_format_sql_no_pygments(no_pygments):
    sql = 'select 1'
    assert format_sql(sql, {}) == sql


def test_format_sql_no_pygments_non_ascii(no_pygments):
    sql = b"select '\xff'"
    formatted = format_sql(sql, {})
    assert formatted.startswith(u"select '")


def test_format_sql_no_pygments_escape_html(no_pygments):
    sql = 'select x < 1'
    formatted = format_sql(sql, {})
    assert not isinstance(formatted, Markup)
    assert Markup('%s') % formatted == 'select x &lt; 1'


@pytest.mark.skipif(not HAVE_PYGMENTS,
                    reason='test requires the "Pygments" library')
def test_format_sql_pygments():
    sql = 'select 1'
    html = format_sql(sql, {})
    assert isinstance(html, Markup)
    assert html.startswith('<div')
    assert 'select' in html
    assert '1' in html


@pytest.mark.skipif(not HAVE_PYGMENTS,
                    reason='test requires the "Pygments" library')
def test_format_sql_pygments_non_ascii():
    sql = b"select 'abc \xff xyz'"
    html = format_sql(sql, {})
    assert isinstance(html, Markup)
    assert html.startswith('<div')
    assert 'select' in html
    assert 'abc' in html
    assert 'xyz' in html



from __future__ import with_statement

import datetime
import logging
try:
    import threading
except ImportError:
    threading = None

from flask_debugtoolbar.panels import DebugPanel
from flask_debugtoolbar.utils import format_fname

_ = lambda x: x


class ThreadTrackingHandler(logging.Handler):
    def __init__(self):
        if threading is None:
            raise NotImplementedError("threading module is not available, \
                the logging panel cannot be used without it")
        logging.Handler.__init__(self)
        self.records = {}  # a dictionary that maps threads to log records

    def emit(self, record):
        self.get_records().append(record)

    def get_records(self, thread=None):
        """
        Returns a list of records for the provided thread, of if none is
        provided, returns a list for the current thread.
        """
        if thread is None:
            thread = threading.currentThread()
        if thread not in self.records:
            self.records[thread] = []
        return self.records[thread]

    def clear_records(self, thread=None):
        if thread is None:
            thread = threading.currentThread()
        if thread in self.records:
            del self.records[thread]


handler = None
_init_lock = threading.Lock()


def _init_once():
    global handler
    if handler is not None:
        return
    with _init_lock:
        if handler is not None:
            return

        # Call werkzeug's internal logging to make sure it gets configured
        # before we add our handler.  Otherwise werkzeug will see our handler
        # and not configure console logging for the request log.
        # Werkzeug's default log level is INFO so this message probably won't
        # be seen.
        try:
            from werkzeug._internal import _log
        except ImportError:
            pass
        else:
            _log('debug', 'Initializing Flask-DebugToolbar log handler')

        handler = ThreadTrackingHandler()
        logging.root.addHandler(handler)


class LoggingPanel(DebugPanel):
    name = 'Logging'
    has_content = True

    def process_request(self, request):
        _init_once()
        handler.clear_records()

    def get_and_delete(self):
        records = handler.get_records()
        handler.clear_records()
        return records

    def nav_title(self):
        return _("Logging")

    def nav_subtitle(self):
        # FIXME l10n: use ngettext
        num_records = len(handler.get_records())
        return '%s message%s' % (num_records, '' if num_records == 1 else 's')

    def title(self):
        return _('Log Messages')

    def url(self):
        return ''

    def content(self):
        records = []
        for record in self.get_and_delete():
            records.append({
                'message': record.getMessage(),
                'time': datetime.datetime.fromtimestamp(record.created),
                'level': record.levelname,
                'file': format_fname(record.pathname),
                'file_long': record.pathname,
                'line': record.lineno,
            })

        context = self.context.copy()
        context.update({'records': records})

        return self.render('panels/logger.html', context)



import sys

import pytest

from flask_debugtoolbar import _printable


def load_app(name):
    app = __import__(name).app
    app.config['TESTING'] = True
    return app.test_client()


def test_basic_app():
    app = load_app('basic_app')
    index = app.get('/')
    assert index.status_code == 200
    assert b'<div id="flDebug"' in index.data


@pytest.mark.skipif(sys.version_info >= (3,),
                    reason='test only applies to Python 2')
def test_printable_unicode():
    class UnicodeRepr(object):
        def __repr__(self):
            return u'\uffff'

    printable = _printable(UnicodeRepr())
    assert "raised UnicodeEncodeError: 'ascii' codec" in printable


@pytest.mark.skipif(sys.version_info >= (3,),
                    reason='test only applies to Python 2')
def test_printable_non_ascii():
    class NonAsciiRepr(object):
        def __repr__(self):
            return 'a\xffb'

    printable = u'%s' % _printable(NonAsciiRepr())
    # should replace \xff with the unicode ? character
    assert printable == u'a\ufffdb'



try:
    import cProfile as profile
except ImportError:
    import profile
import functools
import pstats

from flask import current_app
from flask_debugtoolbar.panels import DebugPanel
from flask_debugtoolbar.utils import format_fname


class ProfilerDebugPanel(DebugPanel):
    """
    Panel that displays the time a response took with cProfile output.
    """
    name = 'Profiler'

    user_activate = True

    def __init__(self, jinja_env, context={}):
        DebugPanel.__init__(self, jinja_env, context=context)
        if current_app.config.get('DEBUG_TB_PROFILER_ENABLED'):
            self.is_active = True

    def has_content(self):
        return bool(self.profiler)

    def process_request(self, request):
        if not self.is_active:
            return

        self.profiler = profile.Profile()
        self.stats = None

    def process_view(self, request, view_func, view_kwargs):
        if self.is_active:
            func = functools.partial(self.profiler.runcall, view_func)
            functools.update_wrapper(func, view_func)
            return func

    def process_response(self, request, response):
        if not self.is_active:
            return False

        if self.profiler is not None:
            self.profiler.disable()
            try:
                stats = pstats.Stats(self.profiler)
            except TypeError:
                self.is_active = False
                return False
            function_calls = []
            for func in stats.sort_stats(1).fcn_list:
                current = {}
                info = stats.stats[func]

                # Number of calls
                if info[0] != info[1]:
                    current['ncalls'] = '%d/%d' % (info[1], info[0])
                else:
                    current['ncalls'] = info[1]

                # Total time
                current['tottime'] = info[2] * 1000

                # Quotient of total time divided by number of calls
                if info[1]:
                    current['percall'] = info[2] * 1000 / info[1]
                else:
                    current['percall'] = 0

                # Cumulative time
                current['cumtime'] = info[3] * 1000

                # Quotient of the cumulative time divded by the number of
                # primitive calls.
                if info[0]:
                    current['percall_cum'] = info[3] * 1000 / info[0]
                else:
                    current['percall_cum'] = 0

                # Filename
                filename = pstats.func_std_string(func)
                current['filename_long'] = filename
                current['filename'] = format_fname(filename)
                function_calls.append(current)

            self.stats = stats
            self.function_calls = function_calls
            # destroy the profiler just in case
        return response

    def title(self):
        if not self.is_active:
            return "Profiler not active"
        return 'View: %.2fms' % (float(self.stats.total_tt)*1000,)

    def nav_title(self):
        return 'Profiler'

    def nav_subtitle(self):
        if not self.is_active:
            return "in-active"
        return 'View: %.2fms' % (float(self.stats.total_tt)*1000,)

    def url(self):
        return ''

    def content(self):
        if not self.is_active:
            return "The profiler is not activated, activate it to use it"

        context = {
            'stats': self.stats,
            'function_calls': self.function_calls,
        }
        return self.render('panels/profiler.html', context)



try:
    import resource
except ImportError:
    pass  # Will fail on Win32 systems
import time
from flask_debugtoolbar.panels import DebugPanel

_ = lambda x: x


class TimerDebugPanel(DebugPanel):
    """
    Panel that displays the time a response took in milliseconds.
    """
    name = 'Timer'
    try:  # if resource module not available, don't show content panel
        resource
    except NameError:
        has_content = False
        has_resource = False
    else:
        has_content = True
        has_resource = True

    def process_request(self, request):
        self._start_time = time.time()
        if self.has_resource:
            self._start_rusage = resource.getrusage(resource.RUSAGE_SELF)

    def process_response(self, request, response):
        self.total_time = (time.time() - self._start_time) * 1000
        if self.has_resource:
            self._end_rusage = resource.getrusage(resource.RUSAGE_SELF)

    def nav_title(self):
        return _('Time')

    def nav_subtitle(self):
        # TODO l10n
        if not self.has_resource:
            return 'TOTAL: %0.2fms' % (self.total_time)

        utime = self._end_rusage.ru_utime - self._start_rusage.ru_utime
        stime = self._end_rusage.ru_stime - self._start_rusage.ru_stime
        return 'CPU: %0.2fms (%0.2fms)' % (
               (utime + stime) * 1000.0, self.total_time)

    def title(self):
        return _('Resource Usage')

    def url(self):
        return ''

    def _elapsed_ru(self, name):
        return (getattr(self._end_rusage, name)
                - getattr(self._start_rusage, name))

    def content(self):

        utime = 1000 * self._elapsed_ru('ru_utime')
        stime = 1000 * self._elapsed_ru('ru_stime')
        vcsw = self._elapsed_ru('ru_nvcsw')
        ivcsw = self._elapsed_ru('ru_nivcsw')
        # minflt = self._elapsed_ru('ru_minflt')
        # majflt = self._elapsed_ru('ru_majflt')

# these are documented as not meaningful under Linux.  If you're running BSD
# feel free to enable them, and add any others that I hadn't gotten to before
# I noticed that I was getting nothing but zeroes and that the docs agreed. :-(
#
#        blkin = self._elapsed_ru('ru_inblock')
#        blkout = self._elapsed_ru('ru_oublock')
#        swap = self._elapsed_ru('ru_nswap')
#        rss = self._end_rusage.ru_maxrss
#        srss = self._end_rusage.ru_ixrss
#        urss = self._end_rusage.ru_idrss
#        usrss = self._end_rusage.ru_isrss

        # TODO l10n on values
        rows = (
            (_('User CPU time'), '%0.3f msec' % utime),
            (_('System CPU time'), '%0.3f msec' % stime),
            (_('Total CPU time'), '%0.3f msec' % (utime + stime)),
            (_('Elapsed time'), '%0.3f msec' % self.total_time),
            (_('Context switches'), '%d voluntary, %d involuntary' % (vcsw, ivcsw)),
            # ('Memory use', '%d max RSS, %d shared, %d unshared' % (rss, srss, urss + usrss)),
            # ('Page faults', '%d no i/o, %d requiring i/o' % (minflt, majflt)),
            # ('Disk operations', '%d in, %d out, %d swapout' % (blkin, blkout, swap)),
        )

        context = self.context.copy()
        context.update({
            'rows': rows,
        })

        return self.render('panels/timer.html', context)



from flask import g
from flask_debugtoolbar.panels import DebugPanel

_ = lambda x: x


class GDebugPanel(DebugPanel):
    """
    A panel to display flask.g content.
    """
    name = 'g'
    has_content = True

    def nav_title(self):
        return _('flask.g')

    def title(self):
        return _('flask.g content')

    def url(self):
        return ''
    
    def content(self):
        context = self.context.copy()
        context.update({
            'g_content': g.__dict__
        })
        return self.render('panels/g.html', context)




import collections
import json
import sys
import uuid

from flask import (
    template_rendered, request, g,
    Response, current_app, abort, url_for
)
from flask_debugtoolbar import module
from flask_debugtoolbar.panels import DebugPanel

_ = lambda x: x


class TemplateDebugPanel(DebugPanel):
    """
    Panel that displays the time a response took in milliseconds.
    """
    name = 'Template'
    has_content = True

    # save the context for the 5 most recent requests
    template_cache = collections.deque(maxlen=5)

    @classmethod
    def get_cache_for_key(self, key):
        for cache_key, value in self.template_cache:
            if key == cache_key:
                return value
        raise KeyError(key)

    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)
        self.key = str(uuid.uuid4())
        self.templates = []
        template_rendered.connect(self._store_template_info)

    def _store_template_info(self, sender, **kwargs):
        # only record in the cache if the editor is enabled and there is
        # actually a template for this request
        if not self.templates and is_editor_enabled():
            self.template_cache.append((self.key, self.templates))
        self.templates.append(kwargs)

    def process_request(self, request):
        pass

    def process_response(self, request, response):
        pass

    def nav_title(self):
        return _('Templates')

    def nav_subtitle(self):
        return "%d rendered" % len(self.templates)

    def title(self):
        return _('Templates')

    def url(self):
        return ''

    def content(self):
        return self.render('panels/template.html', {
            'key': self.key,
            'templates': self.templates,
            'editable': is_editor_enabled(),
        })


def is_editor_enabled():
    return current_app.config.get('DEBUG_TB_TEMPLATE_EDITOR_ENABLED')


def require_enabled():
    if not is_editor_enabled():
        abort(403)


def _get_source(template):
    with open(template.filename, 'rb') as fp:
        source = fp.read()
    return source.decode(_template_encoding())


def _template_encoding():
    return getattr(current_app.jinja_loader, 'encoding', 'utf-8')


@module.route('/template/<key>')
def template_editor(key):
    require_enabled()
    # TODO set up special loader that caches templates it loads
    # and can override template contents
    templates = [t['template'] for t in
                 TemplateDebugPanel.get_cache_for_key(key)]
    return g.debug_toolbar.render('panels/template_editor.html', {
        'static_path': url_for('_debug_toolbar.static', filename=''),
        'request': request,
        'templates': [
            {'name': t.name, 'source': _get_source(t)}
            for t in templates
        ]
    })


@module.route('/template/<key>/save', methods=['POST'])
def save_template(key):
    require_enabled()
    template = TemplateDebugPanel.get_cache_for_key(key)[0]['template']
    content = request.form['content'].encode(_template_encoding())
    with open(template.filename, 'wb') as fp:
        fp.write(content)
    return 'ok'


@module.route('/template/<key>', methods=['POST'])
def template_preview(key):
    require_enabled()
    context = TemplateDebugPanel.get_cache_for_key(key)[0]['context']
    content = request.form['content']
    env = current_app.jinja_env.overlay(autoescape=True)
    try:
        template = env.from_string(content)
        return template.render(context)
    except Exception as e:
        tb = sys.exc_info()[2]
        try:
            while tb.tb_next:
                tb = tb.tb_next
            msg = {'lineno': tb.tb_lineno, 'error': str(e)}
            return Response(json.dumps(msg), status=400,
                            mimetype='application/json')
        finally:
            del tb



try:
    from flask_sqlalchemy import get_debug_queries, SQLAlchemy
except ImportError:
    sqlalchemy_available = False
    get_debug_queries = SQLAlchemy = None
else:
    sqlalchemy_available = True

from flask import request, current_app, abort, g
from flask_debugtoolbar import module
from flask_debugtoolbar.panels import DebugPanel
from flask_debugtoolbar.utils import format_fname, format_sql
import itsdangerous


_ = lambda x: x


def query_signer():
    return itsdangerous.URLSafeSerializer(current_app.config['SECRET_KEY'],
                                          salt='fdt-sql-query')


def is_select(statement):
    prefix = b'select' if isinstance(statement, bytes) else 'select'
    return statement.lower().strip().startswith(prefix)


def dump_query(statement, params):
    if not params or not is_select(statement):
        return None

    try:
        return query_signer().dumps([statement, params])
    except TypeError:
        return None


def load_query(data):
    try:
        statement, params = query_signer().loads(request.args['query'])
    except (itsdangerous.BadSignature, TypeError):
        abort(406)

    # Make sure it is a select statement
    if not is_select(statement):
        abort(406)

    return statement, params


def extension_used():
    return 'sqlalchemy' in current_app.extensions


def recording_enabled():
    return (current_app.debug
            or current_app.config.get('SQLALCHEMY_RECORD_QUERIES'))


def is_available():
    return sqlalchemy_available and extension_used() and recording_enabled()


def get_queries():
    if get_debug_queries:
        return get_debug_queries()
    else:
        return []


class SQLAlchemyDebugPanel(DebugPanel):
    """
    Panel that displays the time a response took in milliseconds.
    """
    name = 'SQLAlchemy'

    @property
    def has_content(self):
        return bool(get_queries()) or not is_available()

    def process_request(self, request):
        pass

    def process_response(self, request, response):
        pass

    def nav_title(self):
        return _('SQLAlchemy')

    def nav_subtitle(self):
        count = len(get_queries())

        if not count and not is_available():
            return 'Unavailable'

        return '%d %s' % (count, 'query' if count == 1 else 'queries')

    def title(self):
        return _('SQLAlchemy queries')

    def url(self):
        return ''

    def content(self):
        queries = get_queries()

        if not queries and not is_available():
            return self.render('panels/sqlalchemy_error.html', {
                'sqlalchemy_available': sqlalchemy_available,
                'extension_used': extension_used(),
                'recording_enabled': recording_enabled(),
            })

        data = []
        for query in queries:
            data.append({
                'duration': query.duration,
                'sql': format_sql(query.statement, query.parameters),
                'signed_query': dump_query(query.statement, query.parameters),
                'context_long': query.context,
                'context': format_fname(query.context)
            })
        return self.render('panels/sqlalchemy.html', {'queries': data})

# Panel views


@module.route('/sqlalchemy/sql_select', methods=['GET', 'POST'])
@module.route('/sqlalchemy/sql_explain', methods=['GET', 'POST'],
              defaults=dict(explain=True))
def sql_select(explain=False):
    statement, params = load_query(request.args['query'])
    engine = SQLAlchemy().get_engine(current_app)

    if explain:
        if engine.driver == 'pysqlite':
            statement = 'EXPLAIN QUERY PLAN\n%s' % statement
        else:
            statement = 'EXPLAIN\n%s' % statement

    result = engine.execute(statement, params)
    return g.debug_toolbar.render('panels/sqlalchemy_select.html', {
        'result': result.fetchall(),
        'headers': result.keys(),
        'sql': format_sql(statement, params),
        'duration': float(request.args['duration']),
    })



import os
import warnings

from flask import Blueprint, current_app, request, g, send_from_directory, url_for
from flask.globals import _request_ctx_stack
from jinja2 import __version__ as __jinja_version__
from jinja2 import Environment, PackageLoader
from werkzeug.urls import url_quote_plus

from flask_debugtoolbar.compat import iteritems
from flask_debugtoolbar.toolbar import DebugToolbar
from flask_debugtoolbar.utils import decode_text, gzip_compress, gzip_decompress

try:
    # Python 3.8+
    from importlib.metadata import version

    __version__ = version("Flask-DebugToolbar")
except ImportError:
    import pkg_resources

    __version__ = pkg_resources.get_distribution("Flask-DebugToolbar").version


module = Blueprint('debugtoolbar', __name__)


def replace_insensitive(string, target, replacement):
    """Similar to string.replace() but is case insensitive
    Code borrowed from:
    http://forums.devshed.com/python-programming-11/case-insensitive-string-replace-490921.html
    """
    no_case = string.lower()
    index = no_case.rfind(target.lower())
    if index >= 0:
        return string[:index] + replacement + string[index + len(target):]
    else:  # no results so return the original string
        return string


def _printable(value):
    try:
        return decode_text(repr(value))
    except Exception as e:
        return '<repr(%s) raised %s: %s>' % (
               object.__repr__(value), type(e).__name__, e)


class DebugToolbarExtension(object):
    _static_dir = os.path.realpath(
        os.path.join(os.path.dirname(__file__), 'static'))

    _redirect_codes = [301, 302, 303, 304]

    def __init__(self, app=None):
        self.app = app
        self.debug_toolbars = {}
        jinja_extensions = ['jinja2.ext.i18n']

        if __jinja_version__[0] == '2':
            jinja_extensions.append('jinja2.ext.with_')

        # Configure jinja for the internal templates and add url rules
        # for static data
        self.jinja_env = Environment(
            autoescape=True,
            extensions=jinja_extensions,
            loader=PackageLoader(__name__, 'templates'))
        self.jinja_env.filters['urlencode'] = url_quote_plus
        self.jinja_env.filters['printable'] = _printable
        self.jinja_env.globals['url_for'] = url_for

        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        for k, v in iteritems(self._default_config(app)):
            app.config.setdefault(k, v)

        if not app.config['DEBUG_TB_ENABLED']:
            return

        if not app.config.get('SECRET_KEY'):
            raise RuntimeError(
                "The Flask-DebugToolbar requires the 'SECRET_KEY' config "
                "var to be set")

        DebugToolbar.load_panels(app)

        app.before_request(self.process_request)
        app.after_request(self.process_response)
        app.teardown_request(self.teardown_request)

        # Monkey-patch the Flask.dispatch_request method
        app.dispatch_request = self.dispatch_request

        app.add_url_rule('/_debug_toolbar/static/<path:filename>',
                         '_debug_toolbar.static', self.send_static_file)

        app.register_blueprint(module, url_prefix='/_debug_toolbar/views')

    def _default_config(self, app):
        return {
            'DEBUG_TB_ENABLED': app.debug,
            'DEBUG_TB_HOSTS': (),
            'DEBUG_TB_INTERCEPT_REDIRECTS': True,
            'DEBUG_TB_PANELS': (
                'flask_debugtoolbar.panels.versions.VersionDebugPanel',
                'flask_debugtoolbar.panels.timer.TimerDebugPanel',
                'flask_debugtoolbar.panels.headers.HeaderDebugPanel',
                'flask_debugtoolbar.panels.request_vars.RequestVarsDebugPanel',
                'flask_debugtoolbar.panels.config_vars.ConfigVarsDebugPanel',
                'flask_debugtoolbar.panels.template.TemplateDebugPanel',
                'flask_debugtoolbar.panels.sqlalchemy.SQLAlchemyDebugPanel',
                'flask_debugtoolbar.panels.logger.LoggingPanel',
                'flask_debugtoolbar.panels.route_list.RouteListDebugPanel',
                'flask_debugtoolbar.panels.profiler.ProfilerDebugPanel',
                'flask_debugtoolbar.panels.g.GDebugPanel',
            ),
        }

    def dispatch_request(self):
        """Modified version of Flask.dispatch_request to call process_view."""
        req = _request_ctx_stack.top.request
        app = current_app

        if req.routing_exception is not None:
            app.raise_routing_exception(req)

        rule = req.url_rule

        # if we provide automatic options for this URL and the
        # request came with the OPTIONS method, reply automatically
        if getattr(rule, 'provide_automatic_options', False) \
           and req.method == 'OPTIONS':
            return app.make_default_options_response()

        # otherwise dispatch to the handler for that endpoint
        view_func = app.view_functions[rule.endpoint]
        view_func = self.process_view(app, view_func, req.view_args)

        return view_func(**req.view_args)

    def _show_toolbar(self):
        """Return a boolean to indicate if we need to show the toolbar."""
        if request.blueprint == 'debugtoolbar':
            return False

        hosts = current_app.config['DEBUG_TB_HOSTS']
        if hosts and request.remote_addr not in hosts:
            return False

        return True

    def send_static_file(self, filename):
        """Send a static file from the flask-debugtoolbar static directory."""
        return send_from_directory(self._static_dir, filename)

    def process_request(self):
        g.debug_toolbar = self

        if not self._show_toolbar():
            return

        real_request = request._get_current_object()

        self.debug_toolbars[real_request] = (
            DebugToolbar(real_request, self.jinja_env))

        for panel in self.debug_toolbars[real_request].panels:
            panel.process_request(real_request)

    def process_view(self, app, view_func, view_kwargs):
        """ This method is called just before the flask view is called.
        This is done by the dispatch_request method.
        """
        real_request = request._get_current_object()
        try:
            toolbar = self.debug_toolbars[real_request]
        except KeyError:
            return view_func

        for panel in toolbar.panels:
            new_view = panel.process_view(real_request, view_func, view_kwargs)
            if new_view:
                view_func = new_view

        return view_func

    def process_response(self, response):
        real_request = request._get_current_object()
        if real_request not in self.debug_toolbars:
            return response

        # Intercept http redirect codes and display an html page with a
        # link to the target.
        if current_app.config['DEBUG_TB_INTERCEPT_REDIRECTS']:
            if response.status_code in self._redirect_codes:
                redirect_to = response.location
                redirect_code = response.status_code
                if redirect_to:
                    content = self.render('redirect.html', {
                        'redirect_to': redirect_to,
                        'redirect_code': redirect_code
                    })
                    response.content_length = len(content)
                    response.location = None
                    response.response = [content]
                    response.status_code = 200

        # If the http response code is 200 then we process to add the
        # toolbar to the returned html response.
        if not (response.status_code == 200 and
                response.is_sequence and
                response.headers['content-type'].startswith('text/html')):
            return response

        if 'gzip' in response.headers.get('Content-Encoding', ''):
            response_html = gzip_decompress(response.data).decode(response.charset)
        else:
            response_html = response.data.decode(response.charset)

        no_case = response_html.lower()
        body_end = no_case.rfind('</body>')

        if body_end >= 0:
            before = response_html[:body_end]
            after = response_html[body_end:]
        elif no_case.startswith('<!doctype html>'):
            before = response_html
            after = ''
        else:
            warnings.warn('Could not insert debug toolbar.'
                          ' </body> tag not found in response.')
            return response

        toolbar = self.debug_toolbars[real_request]

        for panel in toolbar.panels:
            panel.process_response(real_request, response)

        toolbar_html = toolbar.render_toolbar()

        content = ''.join((before, toolbar_html, after))
        content = content.encode(response.charset)
        if 'gzip' in response.headers.get('Content-Encoding', ''):
            content = gzip_compress(content)
        response.response = [content]
        response.content_length = len(content)

        return response

    def teardown_request(self, exc):
        self.debug_toolbars.pop(request._get_current_object(), None)

    def render(self, template_name, context):
        template = self.jinja_env.get_template(template_name)
        return template.render(**context)



from flask_debugtoolbar.panels import DebugPanel

_ = lambda x: x


class HeaderDebugPanel(DebugPanel):
    """
    A panel to display HTTP headers.
    """
    name = 'Header'
    has_content = True
    # List of headers we want to display
    header_filter = (
        'CONTENT_TYPE',
        'HTTP_ACCEPT',
        'HTTP_ACCEPT_CHARSET',
        'HTTP_ACCEPT_ENCODING',
        'HTTP_ACCEPT_LANGUAGE',
        'HTTP_CACHE_CONTROL',
        'HTTP_CONNECTION',
        'HTTP_HOST',
        'HTTP_KEEP_ALIVE',
        'HTTP_REFERER',
        'HTTP_USER_AGENT',
        'QUERY_STRING',
        'REMOTE_ADDR',
        'REMOTE_HOST',
        'REQUEST_METHOD',
        'SCRIPT_NAME',
        'SERVER_NAME',
        'SERVER_PORT',
        'SERVER_PROTOCOL',
        'SERVER_SOFTWARE',
    )

    def nav_title(self):
        return _('HTTP Headers')

    def title(self):
        return _('HTTP Headers')

    def url(self):
        return ''

    def process_request(self, request):
        self.headers = dict(
            [(k, request.environ[k])
                for k in self.header_filter if k in request.environ]
        )

    def content(self):
        context = self.context.copy()
        context.update({
            'headers': self.headers
        })
        return self.render('panels/headers.html', context)



"""Base DebugPanel class"""


class DebugPanel(object):
    """
    Base class for debug panels.
    """
    # name = Base

    # If content returns something, set to true in subclass
    has_content = False

    # If the client is able to activate/de-activate the panel
    user_enable = False

    # We'll maintain a local context instance so we can expose our template
    # context variables to panels which need them:
    context = {}

    # Panel methods
    def __init__(self, jinja_env, context={}):
        self.context.update(context)
        self.jinja_env = jinja_env

        # If the client enabled the panel
        self.is_active = False

    def render(self, template_name, context):
        template = self.jinja_env.get_template(template_name)
        return template.render(**context)

    def dom_id(self):
        return 'flDebug%sPanel' % (self.name.replace(' ', ''))

    def nav_title(self):
        """Title showing in toolbar"""
        raise NotImplementedError

    def nav_subtitle(self):
        """Subtitle showing until title in toolbar"""
        return ''

    def title(self):
        """Title showing in panel"""
        raise NotImplementedError

    def url(self):
        raise NotImplementedError

    def content(self):
        raise NotImplementedError

    # Standard middleware methods
    def process_request(self, request):
        pass

    def process_view(self, request, view_func, view_kwargs):
        pass

    def process_response(self, request, response):
        pass



try:
    from urllib.parse import unquote
except ImportError:
    from urllib import unquote

from flask import url_for, current_app
from werkzeug.utils import import_string


class DebugToolbar(object):

    _cached_panel_classes = {}

    def __init__(self, request, jinja_env):
        self.jinja_env = jinja_env
        self.request = request
        self.panels = []

        self.template_context = {
            'static_path': url_for('_debug_toolbar.static', filename='')
        }

        self.create_panels()

    def create_panels(self):
        """
        Populate debug panels
        """
        activated = self.request.cookies.get('fldt_active', '')
        activated = unquote(activated).split(';')

        for panel_class in self._iter_panels(current_app):
            panel_instance = panel_class(jinja_env=self.jinja_env,
                                         context=self.template_context)

            if panel_instance.dom_id() in activated:
                panel_instance.is_active = True

            self.panels.append(panel_instance)

    def render_toolbar(self):
        context = self.template_context.copy()
        context.update({'panels': self.panels})

        template = self.jinja_env.get_template('base.html')
        return template.render(**context)

    @classmethod
    def load_panels(cls, app):
        for panel_class in cls._iter_panels(app):
            # just loop to make sure they've been loaded
            pass

    @classmethod
    def _iter_panels(cls, app):
        for panel_path in app.config['DEBUG_TB_PANELS']:
            panel_class = cls._import_panel(app, panel_path)
            if panel_class is not None:
                yield panel_class

    @classmethod
    def _import_panel(cls, app, path):
        cache = cls._cached_panel_classes

        try:
            return cache[path]
        except KeyError:
            pass

        try:
            panel_class = import_string(path)
        except ImportError as e:
            app.logger.warning('Disabled %s due to ImportError: %s', path, e)
            panel_class = None

        cache[path] = panel_class
        return panel_class



from setuptools import setup

# Metadata goes in setup.cfg. These are here for GitHub's dependency graph.
setup(
    name="Flask-DebugToolbar",
    install_requires=[
        'Flask>=0.8',
        'Blinker',
        'itsdangerous',
        'werkzeug',
    ],
)



import os
from distutils.sysconfig import get_python_lib

from flask import __version__ as flask_version
from flask_debugtoolbar.panels import DebugPanel

_ = lambda x: x


def relpath(location, python_lib):
    location = os.path.normpath(location)
    relative = os.path.relpath(location, python_lib)
    if relative == os.path.curdir:
        return ''
    elif relative.startswith(os.path.pardir):
        return location
    return relative


class VersionDebugPanel(DebugPanel):
    """
    Panel that displays the Flask version.
    """
    name = 'Version'
    has_content = True

    def nav_title(self):
        return _('Versions')

    def nav_subtitle(self):
        return 'Flask %s' % flask_version

    def url(self):
        return ''

    def title(self):
        return _('Versions')

    def content(self):
        try:
            import pkg_resources
        except ImportError:
            packages = []
        else:
            packages = sorted(pkg_resources.working_set,
                              key=lambda p: p.project_name.lower())

        return self.render('panels/versions.html', {
            'packages': packages,
            'python_lib': os.path.normpath(get_python_lib()),
            'relpath': relpath,
        })



from flask import current_app
from flask_debugtoolbar.panels import DebugPanel

_ = lambda x: x


class ConfigVarsDebugPanel(DebugPanel):
    """
    A panel to display all variables from Flask configuration
    """
    name = 'ConfigVars'
    has_content = True

    def nav_title(self):
        return _('Config')

    def title(self):
        return _('Config')

    def url(self):
        return ''

    def content(self):
        context = self.context.copy()
        context.update({
            'config': current_app.config,
        })

        return self.render('panels/config_vars.html', context)



import itertools
import os.path
import sys
import io
import gzip

try:
    from pygments import highlight
    from pygments.formatters import HtmlFormatter
    from pygments.lexers import SqlLexer
    from pygments.styles import get_style_by_name
    PYGMENT_STYLE = get_style_by_name('colorful')
    HAVE_PYGMENTS = True
except ImportError:
    HAVE_PYGMENTS = False

try:
    import sqlparse
    HAVE_SQLPARSE = True
except ImportError:
    HAVE_SQLPARSE = False

from flask import current_app, Markup


def format_fname(value):
    # If the value has a builtin prefix, return it unchanged
    if value.startswith(('{', '<')):
        return value

    value = os.path.normpath(value)

    # If the file is absolute, try normalizing it relative to the project root
    # to handle it as a project file
    if os.path.isabs(value):
        value = _shortest_relative_path(
            value, [current_app.root_path], os.path)

    # If the value is a relative path, it is a project file
    if not os.path.isabs(value):
        return os.path.join('.', value)

    # Otherwise, normalize other paths relative to sys.path
    return '<%s>' % _shortest_relative_path(value, sys.path, os.path)


def _shortest_relative_path(value, paths, path_module):
    relpaths = _relative_paths(value, paths, path_module)
    return min(itertools.chain(relpaths, [value]), key=len)


def _relative_paths(value, paths, path_module):
    for path in paths:
        try:
            relval = path_module.relpath(value, path)
        except ValueError:
            # on Windows, relpath throws a ValueError for
            # paths with different drives
            continue
        if not relval.startswith(path_module.pardir):
            yield relval


def decode_text(value):
    """
        Decode a text-like value for display.

        Unicode values are returned unchanged. Byte strings will be decoded
        with a text-safe replacement for unrecognized characters.
    """
    if isinstance(value, bytes):
        return value.decode('ascii', 'replace')
    else:
        return value


def format_sql(query, args):
    if HAVE_SQLPARSE:
        query = sqlparse.format(query, reindent=True, keyword_case='upper')

    if not HAVE_PYGMENTS:
        return decode_text(query)

    return Markup(highlight(
        query,
        SqlLexer(),
        HtmlFormatter(noclasses=True, style=PYGMENT_STYLE)))

def gzip_compress(data, compresslevel=6):
    buff = io.BytesIO()
    with gzip.GzipFile(fileobj=buff, mode='wb', compresslevel=compresslevel) as f:
        f.write(data)
    return buff.getvalue()


def gzip_decompress(data):
    with gzip.GzipFile(fileobj=io.BytesIO(data), mode='rb') as f:
        return f.read()
